template<class T> inline Print &operator <<(Print &obj, T arg) { obj.print(arg); return obj; } 
enum _EndLineCode { endl };
inline Print &operator <<(Print &obj, _EndLineCode arg) { obj.println(); return obj; }

//#include <avr/pgmspace.h>
#include <EEPROM.h>
#include <LiquidCrystal.h>

/*=====================================================================================================================
 Morsekeyer by Ingo Strecker, DL6IS
 2014-01-24: Probleme mit der "5" gelöst (hoffentlich)
 2014-01-17: Quittungston für Paddlewechsel; Leerzeichen erscheint statt "5"
 2014-12-29: Paddle-Belegung invertieren; Startnachricht hinzu; Konfigurationsbereich hinzu
 2014-12-11: Umstellung auf Platinenlayout mit 6+2 Textspeichern
 created 2012-04-25
 ======================================================================================================================*/
 
//#####################################################################################################################
//Konfigurationsbereich. Hier dürfen Parameter angepasst werden. Und nur hier!

const byte lcdColumns = 24;                   // Anzahl Zeichen des LCDs pro Zeile (min. 24) 

//Ende des Konfigurationsbereichs.
//#####################################################################################################################*/

/*
 1 WpM =   5 BpM = 1200 ms
 5 WpM =  25 BpM =  240 ms
 10 WpM =  50 BpM =  120 ms
 12 WpM =  60 BpM =  100 ms
 20 WpM = 100 BpM =   60 ms
 50 WpM = 250 BpM =   24 ms
 
 Folgende Angaben stammen aus Wikipedia: Morsecode
 Anfänger kommen kaum über 5 WPM hinaus
 Die Prüfungsgeschwindigkeit für Funkamateure betrug 12 WPM
 Mit viel Übung überschreitet man die 20 WPM-Marke.
 Sehr gute Funker schaffen über 50 WPM
 Der Weltrekord (Mitschrift von 5-er Buchstabengruppen) liegt bei 75,1 WPM
 75,1 WpM = 375,5 BpM = 16 ms
 */


// Nachbildung der ASCII-Tabelle mit Morsezeichen    
const byte ASCIILookup[57] = {
  B1011110, // ' Apostroph
  B101101,  // (
  B1101101, // )
  B1,	    // * kein MorseLookup vorhanden
  B101010,  // +
  B1110011, // Komma
  B1100001, // - Bindestrich
  B1101010, // . Punkt
  B101001,  // /
  B111111,  // 0
  B111110,  // 1
  B111100,  // 2
  B111000,  // 3
  B110000,  // 4
  B100000,  // 5
  B100001,  // 6
  B100011,  // 7
  B100111,  // 8
  B101111,  // 9
  B1000111, // : Doppelpunkt
  B1010101, // ;
  B1,	    // < kein MorseLookup vorhanden
  B110001,  // = gleich
  B1,	    // > kein MorseLookup vorhanden
  B1001100, // ?
  B1010110, // @
  B110,	    // A
  B10001,   // B
  B10101,   // C
  B1001,    // D
  B10,	    // E
  B10100,   // F
  B1011,    // G
  B10000,   // H
  B100,	    // I
  B11110,   // J
  B1101,    // K
  B10010,   // L
  B111,	    // M
  B101,	    // N
  B1111,    // O
  B10110,   // P
  B11011,   // Q
  B1010,    // R
  B1000,    // S
  B11,	    // T
  B1100,    // U
  B11000,   // V
  B1110,    // W
  B11001,   // X
  B11101,   // Y
  B10011,   // Z
  B1,	    // [ kein MorseLookup vorhanden
  B1,	    // \ kein MorseLookup vorhanden
  B1,	    // ] kein MorseLookup vorhanden
  B1,	    // ^ kein MorseLookup vorhanden
  B1101100  // _
};

// Sonderzeichen
const byte ExtendedASCIILookup[32] = {
  B110110,   // À - 192
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B11010,    // Ä - 196
  B110110,   // Å - 197
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B110010,   // È - 200
  B100100,   // É - 201
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B111011,   // Ñ - 209
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B10111,    // Ö - 214
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B11100,    // Ü - 220
  B1,	     // kein MorseLookup vorhanden
  B1,	     // kein MorseLookup vorhanden
  B10011000  , // ß - 223
};

// Signale
const struct {
  byte MZeichen;
  String Zeichen;
} SignaleLookup[] ={
    {B11111, "CH"},   // CH
    {B110101, "KA"},  // KA - Spruchanfang
    {B101000, "VE"},  // VE - Verstanden
    {B1101000, "SK"},  // SK - Verkehrsende
};

//Custom Character #0 - Ä
/* byte SpecialChar0[8]={
 B01010,
 B00000,
 B01110,
 B10001,
 B10001,
 B11111,
 B10001,
 B00000
}; */

//Custom Character #1 - Ö
/* byte SpecialChar1[8]={
 B01010,
 B00000,
 B01110,
 B10001,
 B10001,
 B10001,
 B01110,
 B00000
}; */

//Custom Character #2 - Ü
/* byte SpecialChar2[8]={
 B01010,
 B00000,
 B10001,
 B10001,
 B10001,
 B10001,
 B01110,
 B00000
}; */

//Custom Character #3 - Å
byte SpecialChar3[8]={
 B00100,
 B00000,
 B01110,
 B10001,
 B10001,
 B11111,
 B10001,
 B00000
};

//Custom Character #4 - È
byte SpecialChar4[8]={
 B01000,
 B00100,
 B11111,
 B10000,
 B11110,
 B10000,
 B11111,
 B00000
};

//Custom Character #5 - É
byte SpecialChar5[8]={
 B00010,
 B00100,
 B11111,
 B10000,
 B11110,
 B10000,
 B11111,
 B00000
};

//Custom Character #6 - Ñ
/* byte SpecialChar6[8]={
 B01010,
 B10100,
 B00000,
 B11001,
 B10101,
 B10011,
 B10001,
 B00000
}; */

//Custom Character #7 - ß
/* byte SpecialChar7[8]={
 B01110,
 B10001,
 B10010,
 B10010,
 B10001,
 B10001,
 B10110,
 B00000
}; */

LiquidCrystal lcd (7, 6, 5, 4, 3, 2);

const byte TRXon = 11;     // AnalogButton für die TRX-Tastung
const byte TRXled = 13;    // LED zur Anzeige, ob TRX getastet wird
const byte TRXkey = 15;    // Analog-Port 1 als digitaler Ausgang
const byte PinStrich = 12;
const byte PinPunkt = 11; 
const byte PinSummer = 10;
const byte PinLED = 9;
const byte PinTaster = 8;
byte current, currentA;
byte previous = HIGH, previousA = 0;
byte count, countA, countChar;
unsigned long firstTime = 0, firstTimeA = 0, test = 0;
int WpM;
int Periode; 
int BpM;
String Klartext = "", Code = "";
byte Konsole;
byte useLED = 1;
byte useSummer = 1;
int toneFreq = 800;             // Tonfrequenz für Lautsprecher: 800 Hz
const int memSpeed = 0;         // EEPROM
const int memLED = 1;           // Adressen
const int memSummer = 2;        // für
const int memToneFreq = 3;      // Variablenspeicher
const int memReversePaddle = 4; // und
const int memKey[] = {013, 113, 213, 313, 413, 513, 613, 713, 813, 913};  // Textspeicher
byte anzahlZeichen = 0;
boolean getastet, gefunden = false;
boolean serial = false;
boolean pause = true;
unsigned long ZeitVergangen = 0;  // Zeitverbrauch des Programmablaufs feststellen, um die Gebezeit einzuhalten
int Sonderzeichen;
boolean programming = false;
int x = 0;
byte memText;
byte TextSpeicher[100];
byte TextSpeicherPos = 0;
byte tasterByte = 1;
byte reverseByte = 0;
byte zusZeichen;
boolean memAusgeben = false;
boolean Irrung = false;
byte whichButton;
byte Toleranz = 5;
unsigned long Start = 0;
byte Symbol = 0;
byte TRX = 0;
byte reversePaddle = false;
byte startMessage = true;


//###########################################################
void chooseSpeed() {    // Gebe-Geschwindigkeit einstellen
  delay (200);
  lcd.clear(); 
  lcd.setCursor (0, 0);
  lcd.print (F("Gebe-Geschwindigkeit:"));
  do {
    BpM = WpM * 5;
    Periode = 1200 / WpM;             // Dits pro Minute
    lcd.setCursor (0, 1);
    lcd.print (F("WpM: "));  lcd.print(WpM);  lcd.print(F(", BpM: "));  lcd.print(BpM);  lcd.print (F("   "));
    if (!digitalRead (PinStrich)) {   // Geschwindigkeit langsamer, min. 1
      WpM -= 1;
      if (WpM < 5)
        WpM = 5; 
      delay (200);
    }
    if (!digitalRead (PinPunkt)) {    // Geschwindigkeit schneller, max. 75
      WpM += 1;
      if (WpM >60)
        WpM = 60;
      delay (200);
    }
  } 
  while (digitalRead (PinTaster));
  delay (200);
  EEPROM.write (memSpeed, WpM);
  lcd.clear();
}

//###########################################################
void chooseOutput() {   // Betriebsparameter festlegen
  delay (200);
  lcd.clear(); 
  lcd.setCursor (0, 0);
  lcd.print (F("LED/Summer ein/aus:"));
  do {
    lcd.setCursor (0, 1);
    lcd.print (F("LED: "));  lcd.print (useLED);  lcd.print (F(" <-|-> Summer: ")); lcd.print (useSummer);
    if (!digitalRead (PinStrich)) {
      useLED = useLED ^ 1;
      delay (200);
    }
    if (!digitalRead (PinPunkt)) {
      useSummer = useSummer ^ 1;
      delay (200);
    }
  } 
  while (digitalRead (PinTaster));
  delay (200);
  EEPROM.write (memLED, useLED);
  EEPROM.write (memSummer, useSummer);

  lcd.clear();
  lcd.setCursor (0, 0);
  lcd.print (F("Tonfrequenz:"));
  do {
    lcd.setCursor (0, 1);
    lcd.print (toneFreq);  lcd.print (F(" Hz"));
    if (!digitalRead (PinStrich)) {
      toneFreq -= 20;
      if (toneFreq < 500)
        toneFreq = 500; 
      delay (200);
    }
    if (!digitalRead (PinPunkt)) {
      toneFreq += 20;
      if (toneFreq >1100)
        toneFreq = 1100;
      delay (200);
    }
  } 
  while (digitalRead (PinTaster));
  delay (200);
  EEPROM.write (memToneFreq, toneFreq / 10);
  lcd.clear();
}

//###########################################################
void displayMorsezeichen() {     // Morszeichen und Klartext auf dem Display ausgeben
  if (Code.length() > lcdColumns) {
    while (Code.length() > lcdColumns) {
      Code = Code.substring(1);  // ähnlich scrollDisplayLeft(), aber ohne Display-Überlauf
      Klartext = Klartext.substring(1);
    }
  }
  lcd.setCursor (0, 0);
  lcd.print (Klartext);
  lcd.setCursor (0, 1);
  lcd.print (Code);
}

//###########################################################
void Morsen (byte Anzahl) {
  unsigned long timeLag = 0;
  switch (Anzahl) {
    case 1:
      Code += ".";
      tasterByte <<= 1;
      break;
    case 3:
      Code += "-";
      reverseByte += tasterByte;
      tasterByte <<= 1;
      break;
  }
  timeLag = millis();
  if (startMessage) {
    startMessage = false;
    lcd.clear();
  }
  displayMorsezeichen();
  timeLag = millis() - timeLag;
//  Serial.println (millis() - test - timeLag);
  test = millis() - timeLag;
  if ((Anzahl == 1) || (Anzahl == 3)) {
    ZeitVergangen = millis();
    //if (useSummer) digitalWrite (PinSummer, HIGH);  // für Piezo-Summer mit Elektronik
    if (useSummer)
      tone (PinSummer, toneFreq);                     // für Lautsprecher
    if (useLED)
      digitalWrite (PinLED, HIGH);
    if (TRX)
      digitalWrite (TRXkey, HIGH);
    delay (Periode * Anzahl);                         // Morse-Symbol ausgeben
    //digitalWrite (PinSummer, LOW);                  // für Piezo-Summer mit Elektronik
    noTone(PinSummer);                                // für Lautsprecher
    digitalWrite (PinLED, LOW);
    digitalWrite (TRXkey, LOW);
//    Serial << ("Periode: ") << Periode << (" timeLag: ") << timeLag << endl;
    delay (Periode - timeLag);                        // Pause nach Symbol, 1 Dit
    if (getastet) anzahlZeichen++;
  }
}

//###########################################################
void PauseZeichen() {
  delay (Periode * 2);    // Pause zwischen Buchstaben, 3 Dit's (1 Dit bereits gegeben)
}

//###########################################################
//void PauseWort() {
//  delay (Periode * 6);    // Pause zwischen Worte, 7 Dit's (1 Dit bereits gegeben)
//}

//###########################################################
void MorseZeichen (byte Zeichen) {  // Eingabe vom Computer auswerten
  byte shiftZeichen;
  if (Zeichen != 254) {  // kein Leerzeichen; Leerzeichen von Wert 32 auf 254 abändern, da B100000 = 32 ist, aber die Zahl 5 darstellt!
    shiftZeichen = Zeichen;
    getastet = true;
    while (shiftZeichen > 1) {      // Wenn der Wert 1 ist, wurde das Zeichen komplett abgearbeitet
      if (shiftZeichen & 1 == 1) {  // Ist Bit 0 gesetzt...
        Symbol = 3;
        Morsen (Symbol);                 // ...dann ist es ein Dah
      } else {
        Symbol = 1;
        Morsen (Symbol);                 // ...ansonsten ein Dit
      }
      shiftZeichen >>= 1;           // Bit shift um 1 nach rechts, um das nächste Bit auswerten zu können
    }
    if (programming) {
      TextSpeicher[TextSpeicherPos] = Zeichen;
      TextSpeicherPos++;
      checkTextLenght();
    }
  } else {
    delay (Periode * 3);
    if (programming) {
      TextSpeicher[TextSpeicherPos] = 254;
      TextSpeicherPos++;
      checkTextLenght();
    }
  }
}

//###########################################################
int ErmittleSonderzeichenLCD (int Zeichen) {
  switch (Zeichen) {
    case 192: // À - 183
      Sonderzeichen = 3;
      break;
    case 196: // Ä - 142
      Sonderzeichen = 225;
      break;
    case 197: // Å - 143
      Sonderzeichen = 3;
      break;
    case 200: // È - 212
      Sonderzeichen = 4;
      break;
    case 201: // É - 144
      Sonderzeichen = 5;
      break;
    case 209: // Ñ - 165
      Sonderzeichen = 238;
      break;
    case 214: // Ö - 153
      Sonderzeichen = 239;
      break;
    case 220: // Ü - 154
      Sonderzeichen = 245;
      break;
    case 223: // ß - 225
      Sonderzeichen = 226;
      break;
    default:
      Sonderzeichen = 32;
  }
  delay(1);  // keine Ahnung, warum das sein muss. Ansonsten kommt É statt È. Komisch.
  return Sonderzeichen;
}

//###########################################################
void checkTextLenght() {
  if (TextSpeicherPos >= 89) {  // Warnton, wenn nur noch 10 Zeichen übrig sind
    tone (PinSummer, 500);
    delay (100);
    noTone (PinSummer);
  }
  if (TextSpeicherPos >= 99) {  // zus. Warnton, wenn keine Zeichen mehr übrig
    tone (PinSummer, 250);
    delay (300);
    noTone (PinSummer);
    TextSpeicherPos--;
  }
}

//###########################################################
byte whichButtonPressed (byte aKey) {
  //Serial << ("Tastenwert: ") << aKey << endl;
  byte Key = 0;
// Tastenzuordnung vom Prototyp Ingo
/*  if (aKey > 17 - Toleranz && aKey < 17 + Toleranz)
    Key = 1;
  if (aKey > 35 - Toleranz && aKey < 35 + Toleranz)
    Key = 2;
  if (aKey > 53 - Toleranz && aKey < 53 + Toleranz)
    Key = 3;
  if (aKey > 70 - Toleranz && aKey < 70 + Toleranz)
    Key = 4;
  if (aKey > 87 - Toleranz && aKey < 87 + Toleranz)
    Key = 5;
  if (aKey > 103 - Toleranz && aKey < 103 + Toleranz)
    Key = 6;
  if (aKey > 119 - Toleranz && aKey < 119 + Toleranz)
    Key = TRXon; */
// Tastenzuordnung vom Arduino Mini. Die Variable Key entspricht der Taste des Shields.
// z.B: Key = 3 entspricht S3; Key = 10 entspricht P-Wechsel, momentan ein Dummywert, der sich noch ändern wird!
  if (aKey > 16 - Toleranz && aKey < 16 + Toleranz)
    Key = TRXon;
  if (aKey > 33 - Toleranz && aKey < 33 + Toleranz) {
    reversePaddle = reversePaddle ^ 1;  //P-Wechsel
    EEPROM.write(memReversePaddle, reversePaddle);
    Key = 0;
    tone (PinSummer, 440, 50);
  }
  if (aKey > 50 - Toleranz && aKey < 50 + Toleranz)
    Key = 8;
  if (aKey > 66 - Toleranz && aKey < 66 + Toleranz)
    Key = 7;
  if (aKey > 82 - Toleranz && aKey < 82 + Toleranz)
    Key = 6;
  if (aKey > 97 - Toleranz && aKey < 97 + Toleranz)
    Key = 5;
  if (aKey > 112 - Toleranz && aKey < 112 + Toleranz)
    Key = 4;
  if (aKey > 127 - Toleranz && aKey < 127 + Toleranz)
    Key = 3;
  if (aKey > 141 - Toleranz && aKey < 141 + Toleranz)
    Key = 2;
  if (aKey > 155 - Toleranz && aKey < 155 + Toleranz)
    Key = 1;
  return Key;
}

//###########################################################
void setup() {
  //lcd.createChar(0, SpecialChar0);  // 8 Sonderzeichen in das LCD laden
  //lcd.createChar(1, SpecialChar1);
  //lcd.createChar(2, SpecialChar2);
  lcd.createChar(3, SpecialChar3);
  lcd.createChar(4, SpecialChar4);
  lcd.createChar(5, SpecialChar5);
  //lcd.createChar(6, SpecialChar6);
  //lcd.createChar(7, SpecialChar7);
  lcd.begin (lcdColumns, 2);        // LCD initialisieren
  lcd.write("     Morsekeyer 2.0     ");
  lcd.setCursor(0, 1);
  lcd.write("       von DL6IS        ");
  Serial.begin (115200);              // serielle Schnittstelle aktivieren
  pinMode (PinPunkt, INPUT);
  digitalWrite (PinPunkt, HIGH);    //Pullup-Widerstand aktivieren
  pinMode (PinStrich, INPUT);
  digitalWrite (PinStrich, HIGH);   //Pullup-Widerstand aktivieren
  pinMode (PinTaster, INPUT);
  digitalWrite (PinTaster, HIGH);   //Pullup-Widerstand aktivieren
  pinMode (PinSummer, OUTPUT);
  pinMode (PinLED, OUTPUT);
  pinMode (TRXled, OUTPUT);
  pinMode (TRXkey, OUTPUT);
  // Variablen aus EEPROM lesen -  Schreibzugriffe max. 100.000 laut Spezifikation, also aufpassen ;-)
  WpM = EEPROM.read (memSpeed);      // Geschwindigkeit festlegen
  if (WpM == 255) WpM = 20;          // falls noch nicht gespeichert, Standardgeschwindigkeit
  useLED = EEPROM.read (memLED);     // LED ja/nein
  if (useLED == 255) useLED = 1;     // falls noch nicht gespeichert, LED ein
  useSummer = EEPROM.read (memSummer);  // Summer ja/nein
  if (useSummer == 255) useSummer = 1;  // falls noch nicht gespeichert, Summer ein
  toneFreq = EEPROM.read (memToneFreq); // Tonfrequenz festlegen
  if (toneFreq == 255) toneFreq = 800;  // falls noch nicht gespeichert, 800 Hz
  else toneFreq *= 10;
  reversePaddle = EEPROM.read(memReversePaddle);  // Paddle-Belegung festlegen
  if (reversePaddle == 255) reversePaddle = 0;
  BpM = WpM * 5;
  Periode = 1200 / WpM;              // Dits pro Minute
  Code = "";
  Klartext = "";
  Start = millis();
}

//###########################################################
void loop() {
  char buffer[10];

  if (!digitalRead (PinPunkt)) {      // Paddle-Punkt
    getastet = true;
    if (!reversePaddle)                  // Tastenwechsel
      Symbol = 1;
    else
      Symbol = 3;
    Morsen (Symbol);
  }
  if (!digitalRead (PinStrich)) {     // Paddle-Strich
    getastet = true;
    if (!reversePaddle)                  // Tastenwechsel
      Symbol = 3;
    else
      Symbol = 1;
    Morsen (Symbol);
  }

  // Tasten-Abfrage: http://arduino.cc/playground/Code/HoldButton
  int currentA = analogRead (0);
  if (currentA > 0 && previousA == 0 && millis() - firstTimeA > 200) {
    firstTimeA = millis();             // wenn Taste gedrückt, Zeit merken
    whichButton = whichButtonPressed (currentA);
  }
  if (currentA > 0 && ((millis() - firstTimeA) % 1000) < 20 && millis() - firstTimeA > 500) {
    countA++;                          // Zähler jede Sekunde inkrementieren
    if (countA >= 1 && whichButton != TRXon && whichButton != 0) {
      lcd.clear(); 
      lcd.setCursor (0, 0);
      lcd.print (F("Textspeicher"));
      lcd.setCursor (0, 1);
      lcd.print (F("beschreiben"));
    }
    delay (20);  // Zeit verbraten. Ansonsten wird diese Abfrage 2x durchlaufen
  }
  if (currentA == 0 && previousA > 0 && countA >=0 && countA < 1) {
    if (programming && whichButton != TRXon && whichButton != 0) {  // Text erstellt; speichern in EEPROM
      if (TextSpeicher[TextSpeicherPos-1] == 254)
        TextSpeicherPos--;
      if (TextSpeicherPos >= 99)
        TextSpeicherPos = 99;
      for (int i=0; i < TextSpeicherPos; i++) {
//      	if (TextSpeicher[i] == ' ') {
//      			TextSpeicher[i] = 254;
//      	}
      	EEPROM.write (memKey[whichButton] + i, TextSpeicher[i]);
//        Serial << (i) << ("\t") << TextSpeicher[i] << ("\n");
      }
      EEPROM.write (memKey[whichButton] + TextSpeicherPos, 255);
      lcd.clear(); 
      Code = "";
      Klartext = "";

      for (int i=0; i < 100; i++) {
        TextSpeicher[i] = 254;
      }
      TextSpeicherPos = 0;
      programming = false;
    } else {                           // gespeicherten Text ausgeben
      if (whichButton == TRXon) {         // TRX steuern
        TRX = TRX ^ 1;
        if (TRX)
          digitalWrite (TRXled, HIGH);
        else
          digitalWrite (TRXled, LOW);
      } else {
//        Serial.println ("SpeicherTextAusgeben");
        for (int i=0; i < 100; i++) {
          TextSpeicher[i] = 254;
        }
        memText = EEPROM.read (memKey[whichButton]);
//        Serial.println ("1. "); Serial.println (Konsole);
        x = 0;
        while (memText != 255) {
          TextSpeicher[x] = memText;
//          Serial.println (memText);
          x++;
          memText = EEPROM.read (memKey[whichButton] + x);
        }
        TextSpeicher[x] = 255;
        if (x > 0)
          memAusgeben = true;
        x = 0;
      }
    }
  }
  
  if (currentA == 0 && previousA > 0 && countA >=1 && whichButton != TRXon && whichButton != 0) {  // Taste >=1 Sekunde gedrückt; Textspeicher füllen
//    Serial.println ("SpeicherTextEingeben");
    for (int i=0; i < 100; i++) {
      TextSpeicher[i] = 254;
    }
    TextSpeicherPos = 0;
    programming = true;
    lcd.clear(); 
    Code = "";
    Klartext = "";
  }
  if (currentA == 0) {
    countA = 0;                        // Taste nicht gedrückt, dann Zähler resetten
  }
  previousA = currentA;                 // Status der Taste übernehmen
  // Ende Tasten-Abfrage



  // Tasten-Abfrage: http://arduino.cc/playground/Code/HoldButton
  current = digitalRead (PinTaster);
  if (current == LOW && previous == HIGH && millis() - firstTime > 200) {
    firstTime = millis();             // wenn Taste gedrückt, Zeit merken
    if (count == 0) {
      lcd.clear();  
      lcd.setCursor (0, 0);
      lcd.print (F("Gebe-Geschwindigkeit"));
      lcd.setCursor (0, 1);
      lcd.print (F("festlegen"));
    }
  }
  if (current == LOW && ((millis() - firstTime) % 1000) < 20 && millis() - firstTime > 500) {
    if (count >= 1) {
      lcd.clear(); 
      lcd.setCursor (0, 0);
      lcd.print (F("Betriebsparameter"));
      lcd.setCursor (0, 1);
      lcd.print (F("festlegen"));
    }
    count++;                          // Zähler jede Sekunde inkrementieren
  }
  if (current == HIGH && previous == LOW && count >=0 && count < 1) {
    chooseSpeed();                   // Taste <1 Sekunde gedrückt
  }
  if (current == HIGH && previous == LOW && count >=1) {
    chooseOutput();                   // Taste >=1 Sekunde gedrückt
  }
  if (current == HIGH) {
    count = 0;                        // Taste nicht gedrückt, dann Zähler resetten
  }
  previous = current;                 // Status der Taste übernehmen
  // Ende Tasten-Abfrage

  if (Serial.available()) {        // Zeichen von der seriellen Konsole (Computer) übernehmen
    Start = millis();
    serial = true;
    Konsole = Serial.read();
    //Serial << ("Konsole: ") << Konsole << endl;
    if ((Konsole >= 97) && (Konsole <= 122)) Konsole -= 32; // toUpperCase ASCII
    if ((Konsole >= 224) && (Konsole <= 255)) Konsole -= 32; // toUpperCase Sonder-ASCII
    if (Konsole != ' ') {  // kein Leerzeichen
      if (Konsole < 192) { // Lookup's durchführen
        Konsole = ASCIILookup[Konsole-39];
      } else {
        Konsole = ExtendedASCIILookup[Konsole-192];
      }
      //Serial << ("Konsole: ") << Konsole << endl;
      if (Konsole != 1) {    // MorseLookup vorhanden
        MorseZeichen (Konsole);
        PauseZeichen();
//        Serial << ("Dauer: ") << (millis() - Start) << endl;
//        Start = millis();
      }
    } else {
      MorseZeichen (254);				// Leerzeichen von Wert 32 auf 254 abändern, da B100000 = 32 ist, aber die Zahl 5 darstellt!
      PauseZeichen();
//      Serial << ("Dauer: ") << (millis() - Start) << endl;
//      Start = millis();
    }
  } else serial = false;
  
  if (memAusgeben) {            // gespeicherten Text aus EEPROM ausgeben
  	Serial << TextSpeicher[x] << endl;
//    if (TextSpeicher[x] == ' ') {
//        TextSpeicher[x] = 254;
//    }
  	Serial << TextSpeicher[x] << endl;
    MorseZeichen (TextSpeicher[x]);
//    Serial << ("Zeitvergangen: ") << millis() - ZeitVergangen << endl;
    PauseZeichen();
    x++;
    if (TextSpeicher[x] == 255)
      memAusgeben = false;
  }

  // Buchstaben-Pause; wenn Buchstabe gegeben, Klartext suchen und ausgeben
//  Serial << ("Anzahl Zeichen: ") << anzahlZeichen << endl;
  if ((millis() - ZeitVergangen >= Periode * Symbol + Periode * 2) && (getastet == true) || (anzahlZeichen == 8)) {
    reverseByte += tasterByte;
    if (reverseByte == 0 && tasterByte == 0 && anzahlZeichen == 8) {  // Spezialfall Irrung abfangen
      delay (Periode * 2);
      Klartext += 'H';
      Klartext += 'H';
      zusZeichen = 0;
      gefunden = true;
      reverseByte =B10000000;  // Dirty trick, damit korrekte Anzahl Leerzeichen im Klartext eingefügt werden.
    }
    for (int i=0; i<57; i++) {                          // Ist es ein Standard-ASCII-Zeichen ???
      if (reverseByte == ASCIILookup[i]) {
        Konsole = i + 39;
        Klartext += char(Konsole);
        zusZeichen = 0;
        gefunden = true;
      }
    }
    for (int i=0; i<32; i++) {                          // Ist es ein erweitertes ASCII-Zeichen ???
      if (reverseByte == ExtendedASCIILookup[i]) {      
        Konsole = i + 192;
        Konsole = ErmittleSonderzeichenLCD (Konsole);   
        Klartext += char(Konsole);
        zusZeichen = 0;
        gefunden = true;
        break;
      }
    }
    for (int i=0; i<4; i++) {                          // Ist es ein Signal ???
      if (reverseByte == SignaleLookup[i].MZeichen) {
        Klartext += SignaleLookup[i].Zeichen[0];
        Klartext += SignaleLookup[i].Zeichen[1];
        zusZeichen = 1;
        gefunden = true;
      }
    }
    if (gefunden) {                                     // Zeichen wurde identifiziert
      if (programming && !serial) {
        TextSpeicher[TextSpeicherPos] = reverseByte;
        TextSpeicherPos++;
        checkTextLenght();
      }
      while (reverseByte != 1) {
        reverseByte >>= 1;
        Klartext += ' ';
      }
      Klartext = Klartext.substring (0, Klartext.length() - zusZeichen);
      Code += " ";
    } else {                                            // Zeichen wurde NICHT identifiziert
      Klartext += "# ";
//      Serial << ("anzahlZeichen: ") << anzahlZeichen << ("\n");
      for (byte i=1; i < anzahlZeichen; i++) {
        Klartext += " ";
      }
      Code += " ";
      PauseZeichen();
    }
    anzahlZeichen = 0;
    //for (byte i=0; i < sizeof(Gegeben) - 1; i++) {  // jede Stelle der Variable auf Null
      //Gegeben[i] = '\0';
    //} 
    tasterByte = 1;
    reverseByte = 0;
    getastet = false;
    gefunden = false;
    pause = false;
  }

  // Wort-Pause; nur noch 4 dit's, da Buchstabenpause auch noch da
  if ((millis() - ZeitVergangen >= Periode * Symbol + Periode * 7) && (!pause)) {
    Code += " ";
    Klartext += " ";
    if (programming) {
      TextSpeicherPos++;
      TextSpeicher[TextSpeicherPos] = ' ';
      checkTextLenght();
    }
    displayMorsezeichen();
    pause = true;
  }
}

